/*
 * Copyright 2013 Victor Nazarov <asviraspossible@gmail.com>.
 */
package org.github.sviperll.stream;

import com.github.sviperll.InterruptableConsumer;
import com.github.sviperll.Consumer;
import com.github.sviperll.Generator;
import com.github.sviperll.io.Closeables;
import java.io.Closeable;
import java.util.Collections;
import java.util.Iterator;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.SynchronousQueue;

/**
 *
 * @author Victor Nazarov <asviraspossible@gmail.com>
 */
class GeneratorStreamBehaviour<T> implements StreamBehaviour<T> {
    private final Generator<T> generator;
    public GeneratorStreamBehaviour(Generator<T> generator) {
        this.generator = generator;
    }

    @Override
    public Iterator<T> openIterator() {
        SynchronousQueue<T> queue = new SynchronousQueue<T>();
        Drainer<T> drainer = new Drainer<T>(generator);
        return drainer.createIterator();
    }

    @Override
    public void forEach(final InterruptableConsumer<? super T> consumer) {
        generator.forEach(new Consumer<T>() {
            @Override
            public void accept(T value) {
                consumer.accept(value);
            }
        });
    }

    private static class Drainer<T> {
        private volatile Generator<T> generator;
        private final BlockingQueue<DrainerRequest> requestQueue = new SynchronousQueue<DrainerRequest>();
        private final BlockingQueue<DrainerResponse<T>> responseQueue = new SynchronousQueue<DrainerResponse<T>>();
        public Drainer(Generator<T> generator) {
            this.generator = generator;
        }

        CloseableIterator<T> createIterator() {
            Runnable runnable = new Runnable() {
                @Override
                public void run() {
                    try {
                        generator.forEach(new Consumer<T>() {
                            @Override
                            public void accept(T value) {
                                DrainerRequest request = takeRequest();
                                if (request.isCloseRequest())
                                    throw new IsClosedException();
                                else {
                                    putResponse(DrainerResponse.fetchedValue(value));
                                }
                            }
                        });
                        takeRequest();
                    } catch (IsClosedException ex) {
                    }
                    putResponse(DrainerResponse.<T>isClosed());
                }
            };
            Thread thread = new Thread(runnable);
            thread.start();
            putRequest(DrainerRequest.FETCH_VALUE);
            DrainerResponse<T> response = takeResponse();
            if (response.isClosedResponse())
                return new CloseableIterator<T>(Collections.<T>emptyIterator(), Closeables.empty());
            else {
                DrainerIterator drainerIterator = new DrainerIterator(response.value());
                return new CloseableIterator<T>(drainerIterator, drainerIterator);
            }
        }

        private class DrainerIterator implements Iterator<T>, Closeable {
            private boolean isClosed = false;
            private T nextValue;

            public DrainerIterator(T nextValue) {
                this.nextValue = nextValue;
            }

            @Override
            public void close() {
                if (!isClosed) {
                    putRequest(DrainerRequest.CLOSE);
                    takeResponse();
                }
            }

            @Override
            public boolean hasNext() {
                return !isClosed;
            }

            @Override
            public T next() {
                T result = nextValue;
                putRequest(DrainerRequest.FETCH_VALUE);
                DrainerResponse<T> response = takeResponse();
                if (response.isClosedResponse())
                    isClosed = true;
                else
                    nextValue = response.value();
                return result;
            }

            @Override
            public void remove() {
                throw new UnsupportedOperationException("Not supported.");
            }
        }

        private DrainerRequest takeRequest() {
            for (;;) {
                try {
                    return requestQueue.take();
                } catch (InterruptedException ex) {
                }
            }
        }

        private void putRequest(DrainerRequest request) {
            for (;;) {
                try {
                    requestQueue.put(request);
                    return;
                } catch (InterruptedException ex) {
                }
            }
        }

        private void putResponse(DrainerResponse<T> response) {
            for (;;) {
                try {
                    responseQueue.put(response);
                    return;
                } catch (InterruptedException ex) {
                }
            }
        }

        private DrainerResponse<T> takeResponse() {
            for (;;) {
                try {
                    return responseQueue.take();
                } catch (InterruptedException ex) {
                }
            }
        }
    }

    @SuppressWarnings("serial")
    private static class IsClosedException extends RuntimeException {
    }

    private static enum DrainerRequest {
        CLOSE {
            @Override
            public boolean isCloseRequest() {
                return true;
            }
            @Override
            public boolean isFetchRequest() {
                return false;
            }
        }, FETCH_VALUE {
            @Override
            public boolean isCloseRequest() {
                return false;
            }
            @Override
            public boolean isFetchRequest() {
                return true;
            }
        };

        public abstract boolean isCloseRequest();
        public abstract boolean isFetchRequest();
    }

    private static class DrainerResponse<T> {
        @SuppressWarnings({"rawtypes", "unchecked"})
        private static final DrainerResponse IS_CLOSED = new DrainerResponse(null);
        public static <T> DrainerResponse<T> fetchedValue(T value) {
            return new DrainerResponse<T>(value);
        }

        @SuppressWarnings("unchecked")
        private static <T> DrainerResponse<T> isClosed() {
            return IS_CLOSED;
        }
        private final T value;

        public DrainerResponse(T value) {
            this.value = value;
        }
        public boolean isClosedResponse() {
            return this == IS_CLOSED;
        }
        public boolean isFetchedValueResponse() {
            return this != IS_CLOSED;
        }

        public T value() {
            if (!isFetchedValueResponse())
                throw new IllegalStateException();
            return value;
        }
    }
}
