/*
 * Copyright 2013 Victor Nazarov <asviraspossible@gmail.com>.
 */
package org.github.sviperll.stream;

import com.github.sviperll.InterruptableConsumer;
import com.github.sviperll.BiFunction;
import com.github.sviperll.BinaryOperator;
import com.github.sviperll.Consumer;
import com.github.sviperll.Function;
import com.github.sviperll.Generator;
import com.github.sviperll.OptionalVisitor;
import com.github.sviperll.Predicate;
import com.github.sviperll.Supplier;
import com.github.sviperll.UnaryOperator;
import com.github.sviperll.io.Closeables;
import java.util.Collections;
import java.util.Iterator;

/**
 *
 * @author Victor Nazarov <asviraspossible@gmail.com>
 */
public class Stream<T> {
    public static <T> Stream<T> newInstance(final StreamBehaviour<T> behaviour) {
        return new Stream<T>(new CloseableStreamBehaviour<T>() {
            @Override
            public CloseableIterator<T> openIterator() {
                return new CloseableIterator<T>(behaviour.openIterator(), Closeables.empty());
            }

            @Override
            public void forEach(InterruptableConsumer<? super T> consumer) {
                behaviour.forEach(consumer);
            }
        });
    }

    static <T> Stream<T> newInstance(final CloseableStreamBehaviour<T> behaviour) {
        return new Stream<T>(behaviour);
    }

    public static <T> Stream<T> newInstance(final Generator<T> generator) {
        return Stream.newInstance(new GeneratorStreamBehaviour<T>(generator));
    }

    public static <T> Stream<T> newInstance(final Supplier<? extends CloseableStream<T>> closeableStreamFactory) {
        return Stream.newInstance(new CloseableStreamSupplierBehaviour<T>(closeableStreamFactory));
    }

    public static <T> Stream<T> generate(final Supplier<T> supplier) {
        return Stream.newInstance(new GeneratedStreamBehaviour<T>(supplier));
    }

    public static <T> Stream<T> iterate(final T seed, final UnaryOperator<T> function) {
        return Stream.newInstance(new IteratedStreamBehaviour<T>(seed, function));
    }

    public static <T> Stream<T> flatten(Stream<? extends Stream<T>> stream) {
        return Stream.newInstance(new FlattenedStreamBehaviour<T>(stream));
    }

    public static <T> Stream<T> empty() {
        return Stream.newInstance(new StreamBehaviour<T>() {
            @Override
            public Iterator<T> openIterator() {
                return Collections.emptyIterator();
            }

            @Override
            public void forEach(InterruptableConsumer<? super T> consumer) {
            }
        });
    }

    public static <T> Stream<T> of(final T... elements) {
        return Arrays.asStream(elements);
    }

    private final CloseableStreamBehaviour<T> behaviour;
    private Stream(CloseableStreamBehaviour<T> behaviour) {
        this.behaviour = behaviour;
    }

    public CloseableIterator<T> openIterator() {
        return behaviour.openIterator();
    }

    public void forEach(InterruptableConsumer<? super T> consumer) {
        behaviour.forEach(consumer);
    }

    public void forEach(final Consumer<? super T> consumer) {
        behaviour.forEach(new InterruptableConsumer<T>() {
            @Override
            public void accept(T value) {
                consumer.accept(value);
            }

            @Override
            public boolean needsMore() {
                return true;
            }
        });
    }

    public <U> Stream<U> map(Function<? super T, U> function) {
        return new Stream<U>(new MappedStreamBehaviour<T, U>(behaviour, function));
    }

    public Stream<T> filter(Predicate<? super T> predicate) {
        return new Stream<T>(new FilteredStreamBehaviour<T>(behaviour, predicate));
    }

    public <U> U reduce(final U seed, final BiFunction<U, ? super T, U> function) {
        return collect(Collectors.reducing(seed, function));
    }

    public <R, E extends Exception> R reduce(final BinaryOperator<T> operator, OptionalVisitor<? super T, R, E> visitor) throws E {
        return collect(Collectors.reducing(operator, visitor));
    }

    public <R> Stream<R> flatMap(Function<? super T, ? extends Stream<R>> function) {
        return flatten(map(function));
    }

    public Stream<T> skip(int offset) {
        return new Stream<T>(new SkippedStreamBehaviour<T>(behaviour, offset));
    }

    public Stream<T> limit(int size) {
        return new Stream<T>(new LimitedStreamBehaviour<T>(behaviour, size));
    }

    public <R, E extends Exception> R collect(Supplier<? extends CollectorState<? super T, R, E>> collector) throws E {
        CollectorState<? super T, R, E> state = collector.get();
        behaviour.forEach(state);
        return state.get();
    }

    public <R, E extends Exception> R findFirst(OptionalVisitor<? super T, R, E> visitor) throws E {
        return collect(Collectors.findFirst(visitor));
    }

    public int count() {
        return collect(Collectors.counting());
    }
}
