/*
 * Copyright 2013 Victor Nazarov <asviraspossible@gmail.com>.
 */
package org.github.sviperll.stream;

import com.github.sviperll.InterruptableConsumer;
import com.github.sviperll.Consumer;
import com.github.sviperll.Supplier;
import com.github.sviperll.io.Closeables;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 *
 * @author Victor Nazarov <asviraspossible@gmail.com>
 */
public class StreamableFile {
    private final File file;
    public StreamableFile(File file) {
        this.file = file;
    }

    public File file() {
        return file;
    }

    public StreamableTextFile text(Charset charset) {
        return new StreamableTextFile(file, charset);
    }

    public Stream<Byte> bytes() {
        return Stream.newInstance(new Supplier<CloseableStream<Byte>>() {
            @Override
            public CloseableStream<Byte> get() {
                try {
                    Owned<BufferedInputStream> closeable = Owned.newInstance(new BufferedInputStream(new FileInputStream(file)));
                    try {
                        InputStream inputStream = closeable.value();
                        int firstByte = inputStream.read();
                        if (firstByte < 0)
                            return new CloseableStream<Byte>(Stream.<Byte>empty(), Closeables.empty());
                        else {
                            Stream<Byte> stream = Stream.newInstance(new InputStreamStream((byte)firstByte, inputStream));
                            return new CloseableStream<Byte>(stream, closeable.take());
                        }
                    } finally {
                        closeable.close();
                    }
                } catch (IOException ex) {
                    throw new RuntimeException(ex);
                }
            }
        });
    }

    private static class InputStreamStream implements StreamBehaviour<Byte> {
        private final byte firstByte;
        private final InputStream inputStream;
        public InputStreamStream(byte firstByte, InputStream inputStream) {
            this.firstByte = firstByte;
            this.inputStream = inputStream;
        }

        @Override
        public Iterator<Byte> openIterator() {
            return new ByteInputStreamIterator(firstByte, inputStream);
        }

        @Override
        public void forEach(InterruptableConsumer<? super Byte> consumer) {
            try {
                if (consumer.needsMore()) {
                    consumer.accept(firstByte);
                    int b;
                    while ((b = inputStream.read()) >= 0) {
                        if (!consumer.needsMore())
                            break;
                        consumer.accept((byte)b);
                    }
                }
            } catch (IOException ex) {
                throw new RuntimeException(ex);
            }
        }
    }

    private static class ByteInputStreamIterator implements Iterator<Byte> {
        private final InputStream inputStream;
        private byte nextByte;
        private boolean hasNext = true;

        public ByteInputStreamIterator(byte firstByte, InputStream inputStream) {
            this.nextByte = firstByte;
            this.inputStream = inputStream;
        }

        @Override
        public boolean hasNext() {
            return hasNext;
        }

        @Override
        public Byte next() {
            if (!hasNext)
                throw new NoSuchElementException();
            byte result = nextByte;
            int byteValue;
            try {
                byteValue = inputStream.read();
            } catch (IOException ex) {
                throw new RuntimeException(ex);
            }
            if (byteValue < 0)
                hasNext = false;
            else
                nextByte = (byte)byteValue;
            return result;
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException("Not supported.");
        }

    }
}
